import signal
import rox, gobject
from rox import g, Menu
import top, os, cell_bar
from support import pretty_size

NAME = 0
SHARED = 1
UNSHARED = 2
SWAPPED = 3
PID = 4
STATE = 5
TIME = 6
COLOUR = 7
RSS = 8

uid = os.getuid()

menu = Menu.Menu('main', [
(_('/Pause (STOP)'),	'pause',	None),
(_('/Unpause (CONT)'),	'unpause',	None),
(_('/'),		None,		"<Separator>"),
(_('/Quit (TERM)'),	'quit', 	None),
(_('/Kill (KILL)'),	'kill', 	None),
])

def load_image(img):
	path = os.path.join(rox.app_dir, 'images', img)
	return g.gdk.pixbuf_new_from_file(path)

state_sleep = load_image("sleep.png");
state_run = load_image("run.png");
state_unint = load_image("unint.png");
state_idle = load_image("idle.png");
state_zombie = load_image("zombie.png");
state_stopped = load_image("stopped.png");

state_to_pixbuf = {
	'S': state_sleep,
	'R': state_run,
	'D': state_unint,
	'I': state_idle,
	'Z': state_zombie,
	'T': state_stopped
}

class Task:
	def __init__(self, pid):
		self.pid = pid
		self.kids = []

		try:
			self.rss, self.shared, self.size = top.get_proc_mem(pid)
			self.ppid = top.get_proc_ppid(pid)

			self.real_time = top.get_proc_time(pid)
			self.state, self.uid, command = top.get_proc_state(pid)

			if command.endswith('AppRun') or \
			   command.endswith('AppletRun'):
				command = os.path.dirname(command)

			if command[0] == '/':
				command = os.path.basename(command)

			if len(command) > 40:
				command = command[:38] + '...'
			self.command = command
		except:
			self.command = '(missing)'
			self.rss = self.shared = self.size = 0
			self.ppid = 1
			self.real_time = 0
			self.state = 'Z'
			self.uid = 0

def cmp_task(a, b): return cmp(a.pid, b.pid)

class Processes(g.ScrolledWindow):
	def __init__(self, window, show_all, hide_swap, show_tree):
		g.ScrolledWindow.__init__(self)
		self.set_policy(g.POLICY_NEVER, g.POLICY_AUTOMATIC)
		self.set_shadow_type(g.SHADOW_IN)
		self.unset_flags(g.CAN_FOCUS)

		tree = self.model = g.TreeStore(str,
				float, float, float,
				long, g.gdk.Pixbuf, long,
				g.gdk.Color, str)

		self.view = view = g.TreeView(model = tree)
		view.set_rules_hint(True)
		self.add(view)

		view.insert_column_with_attributes(
			0, _("Name"), g.CellRendererText(),
			text = NAME, foreground_gdk = COLOUR)

		view.insert_column_with_attributes(
			1, _("PID"), g.CellRendererText(),
			text = PID)

		view.insert_column_with_attributes(
			2, _("S"), g.CellRendererPixbuf(),
			pixbuf = STATE)

		view.insert_column_with_attributes(
			3, _("RSS"), g.CellRendererText(),
			text = RSS)

		self.cell_bar = cell_bar.CellRendererBar()
		view.insert_column_with_attributes(
			4, _("Memory usage (shared, unshared, swapped)"),
			self.cell_bar, part1 = 1, part2 = 2, part3 = 3)

		self.show_all = show_all
		self.hide_swap = hide_swap
		self.show_tree = show_tree

		self.update()

		self.timeout = gobject.timeout_add(2000, self.update)

		view.expand_all()

		def clicked(view, event):
			if event.button == 3:
				menu.popup(self, event)
		view.connect('button-press-event', clicked)
		menu.attach(window, self)

	def update(self, *unused):
		tasks = self.build_proc_tree()
		self.max = 0
		self.merge_tree(tasks, None)
		self.cell_bar.max = max(self.max, 1)
		return True

	def build_proc_tree(self):
		tasks = []
		pid_to_task = {}
		pids = top.get_proclist(self.show_all, uid)
		tasks = map(Task, pids)
		for t in tasks: pid_to_task[t.pid] = t

		top_level = []

		for t in tasks:
			t.parent = pid_to_task.get(t.ppid, None)
			if t.parent == t: t.parent = None
			if t.parent is not None and self.show_tree:
				t.parent.kids.append(t)
			else:
				top_level.append(t)

		for t in tasks:
			t.kids.sort(cmp_task)

		top_level.sort(cmp_task)

		return top_level

	def merge_tree(self, tasks, parent):
		model = self.model

		if parent:
			iter = model.iter_children(parent)
		else:
			iter = model.get_iter_first()

		for t in tasks:
			state = state_to_pixbuf.get(t.state, None)

			need_expand = False

			# Delete every process until we get to our one
			# (or go past it)
			while 1:
				if not iter:
					# No more existing tasks. Append new.
					new = model.append(parent)
					need_expand = True
					break
				pid = model[iter][PID]
				time = model[iter][TIME]
				if pid == t.pid:
					new = iter
					if state == state_sleep and t.real_time > time:
						# If it ran since last check
						# then mark it as idle, rather
						# than sleeping.
						state = state_idle
					break
				elif pid > t.pid:
					need_expand = True
					new = model.insert_before(parent, iter)
					break
				else:
					# Delete this entry
					to_delete = iter
					iter = model.iter_next(iter)
					model.remove(to_delete)

			unshared=t.rss - t.shared
			if unshared<0:
				unshared=0
			model.set(new,
				NAME, t.command,
				SHARED, t.shared,
				UNSHARED, unshared,
				SWAPPED, t.size - t.rss,
				PID, t.pid,
				STATE, state,
				TIME, t.real_time,
				RSS, pretty_size(t.rss))

			if t.uid != uid:
				model.set(new, COLOUR,
					self.view.style.fg[g.STATE_INSENSITIVE])

			if self.hide_swap:
				self.max = max(self.max, t.rss)
			else:
				self.max = max(self.max, t.size)

			self.merge_tree(t.kids, new)

			if need_expand:
				path = model.get_path(new)
				self.view.expand_row(path, False)
				path = path[:-1]
				if path:
					self.view.expand_row(path, False)

			iter = model.iter_next(new)

		while iter:
			to_delete = iter
			iter = model.iter_next(iter)
			model.remove(to_delete)

	def get_best_size(self):
		return self.view.size_request()

	def toggle_show_all(self):
		self.show_all = not self.show_all
		self.update()
	
	def toggle_hide_swap(self):
		self.hide_swap = not self.hide_swap
		self.update()
		
	def toggle_show_tree(self):
		self.show_tree = not self.show_tree
		self.update()
		self.view.expand_all()
	
	def signal(self, num):
		sel = self.view.get_selection()
		store, iter = sel.get_selected()
		pid = store[iter][PID]
		os.kill(pid, num)
		self.update()

	def pause(self): self.signal(signal.SIGSTOP)
	def unpause(self): self.signal(signal.SIGCONT)
	def quit(self): self.signal(signal.SIGTERM)
	def kill(self): self.signal(signal.SIGKILL)
