import rox, os

try:
	from ctypes import cdll
except ImportError:
	rox.croak('The ctypes python module could not be imported.\n'
		  'Is it installed? You can get it from here:\n\n'
		  'http://starship.python.net/crew/theller/ctypes/')

from ctypes import c_void_p, Structure, c_longlong, byref, c_char_p
from ctypes import POINTER, pointer, c_char, c_int, c_uint, c_ulonglong


try:
	try:
		libgtop = cdll.LoadLibrary('libgtop-2.0.so.5')
		top_version = (2, 5)
	except OSError:
		libgtop = cdll.LoadLibrary('libgtop-2.0.so')
		top_version = (2, 0)
except:
	rox.report_exception()
	rox.croak('Failed to load gtop2 C library. Is it installed?\n\n'
		  'You can download the library from here:\n'
		  'http://www.gnome.org/softwaremap/projects/libgtop/')

GLIBTOP_MOUNTENTRY_LEN = 79
GLIBTOP_NCPU = 4
GLIBTOP_PROC_STATE_STATE = 1

GLIBTOP_PROC_STATE_CMD = 0
GLIBTOP_PROC_STATE_UID = 2
GLIBTOP_MAX_GROUPS = 64

server = c_void_p.in_dll(libgtop, 'glibtop_global_server')

try:
	glibtop_free_r=libgtop.glibtop_free_r
	version=0
except:
	version=2
	glib=cdll.LoadLibrary('libglib-2.0.so')
	def glibtop_free_r(ignored, ptr):
		glib.g_free(ptr)

if version==2:
	class glibtop_fsusage(Structure):
		_fields_ = [('flags', c_longlong),
			    ('blocks', c_longlong),
			    ('bfree', c_longlong),
			    ('bavail', c_longlong),
			    ('files', c_longlong),
			    ('ffree', c_longlong),
			    ('block_size', c_int),
			    ('read', c_longlong),
			    ('write', c_longlong),
			]

	class glibtop_proc_state(Structure):
		_fields_ = [('flags', c_ulonglong),
			    ('cmd', c_char * 40),
			    ('state', c_uint),
			    ('uid', c_int),
			    ('gid', c_int),
			    ('ruid', c_int),
			    ('rgid', c_int),
			    ('has_cpu', c_int),
			    ('processor', c_int),
			    ('last_processor', c_int),
			]
else:
	class glibtop_fsusage(Structure):
		_fields_ = [('flags', c_longlong),
			    ('blocks', c_longlong),
			    ('bfree', c_longlong),
			    ('bavail', c_longlong),
			    ('files', c_longlong),
			    ('ffree', c_longlong),
			    ]

	class glibtop_proc_state(Structure):
		_fields_ = [('flags', c_ulonglong),
			    ('cmd', c_char * 40),
			    ('state', c_char),
			    ('uid', c_int),
			    ('gid', c_int),
			    ]

class glibtop_mountlist(Structure):
	_fields_ = [('flags', c_longlong),
		    ('number', c_longlong),
		    ('total', c_longlong),
		    ('size', c_longlong)]

class glibtop_mountentry(Structure):
	_fields_ = [('dev', c_longlong),
		    ('devname', c_char * (GLIBTOP_MOUNTENTRY_LEN + 1)),
		    ('mountdir', c_char * (GLIBTOP_MOUNTENTRY_LEN + 1)),
		    ('type', c_char * (GLIBTOP_MOUNTENTRY_LEN + 1))]

class glibtop_proc_time(Structure):
	_fields_ = [('flags', c_ulonglong),
		    ('start_time', c_ulonglong),
		    ('rtime', c_ulonglong),
		    ('utime', c_ulonglong),
		    ('stime', c_ulonglong),
		    ('cutime', c_ulonglong),
		    ('cstime', c_ulonglong),
		    ('timeout', c_ulonglong),
		    ('it_real_value', c_ulonglong),
		    ('frequency', c_ulonglong),
 		    ('xcpu_utime', c_ulonglong * GLIBTOP_NCPU),
 		    ('xcpu_stime', c_ulonglong * GLIBTOP_NCPU)]

if version==2:
	class glibtop_proc_uid(Structure):
		_fields_ = [('flags', c_ulonglong),
			    ('uid',c_int),
			    ('euid',c_int),
			    ('gid',c_int),
			    ('egid',c_int),
			    ('suid',c_int),
			    ('sgid',c_int),
			    ('fsuid',c_int),
			    ('fsgid',c_int),
			    ('pid',c_int),
			    ('ppid',c_int),
			    ('pgrp',c_int),
			    ('session',c_int),
			    ('tty',c_int),
			    ('tpgid',c_int),
			    ('priority',c_int),
			    ('nice', c_int),
			    ('ngroups', c_int),
			    ('groups', c_int * GLIBTOP_MAX_GROUPS),
			]
else:
	class glibtop_proc_uid(Structure):
		_fields_ = [('flags', c_ulonglong),
			    ('uid',c_int),
			    ('euid',c_int),
			    ('gid',c_int),
			    ('egid',c_int),
			    ('pid',c_int),
			    ('ppid',c_int),
			    ('pgrp',c_int),
			    ('session',c_int),
			    ('tty',c_int),
			    ('tpgid',c_int),
			    ('priority',c_int),
			    ('nice', c_int),
			    ('ngroups',c_int),
			    ('nice', c_int * GLIBTOP_MAX_GROUPS),
			    ]

class glibtop_proc_mem(Structure):
	_fields_ = [('flags', c_longlong),
		    ('size', c_longlong),
		    ('vsize',  c_longlong),
		    ('resident',  c_longlong),
		    ('share',  c_longlong),
		    ('rss',  c_longlong),
		    ('rss_rlim', c_longlong)]

class glibtop_proc_args(Structure):
	_fields_ = [('flags', c_ulonglong),
		    ('size', c_ulonglong)]

GLIBTOP_KERN_PROC_ALL = 0
GLIBTOP_KERN_PROC_PID = 1
GLIBTOP_KERN_PROC_UID = 5

if version==2:
	GLIBTOP_PROC_UID_PPID = 9
else:
	GLIBTOP_PROC_UID_PPID = 5

GLIBTOP_PROC_MEM_SIZE = 0
GLIBTOP_PROC_MEM_VSIZE = 1
GLIBTOP_PROC_MEM_RESIDENT = 2
GLIBTOP_PROC_MEM_SHARE = 3
GLIBTOP_PROC_MEM_RSS = 4
GLIBTOP_PROC_MEM_RSS_RLIM = 5

GLIBTOP_MAX_PROC_MEM = 6

if version==2:
	GLIBTOP_PROCESS_RUNNING = 1
	GLIBTOP_PROCESS_INTERRUPTIBLE = 2
	GLIBTOP_PROCESS_UNINTERRUPTIBLE = 4
	GLIBTOP_PROCESS_ZOMBIE = 8
	GLIBTOP_PROCESS_STOPPED = 16
	GLIBTOP_PROCESS_SWAPPING = 32
	GLIBTOP_PROCESS_DEAD = 64	


class glibtop_proclist(Structure):
	_fields_ = [('flags', c_ulonglong),
		    ('number', c_ulonglong),
		    ('total', c_ulonglong),
		    ('size', c_ulonglong)]

libgtop.glibtop_get_mountlist_l.restype = POINTER(glibtop_mountentry)
libgtop.glibtop_get_proclist_l.restype = POINTER(c_uint)
libgtop.glibtop_get_proc_args_l.restype = POINTER(c_char)

def get_mountlist():
 	info = glibtop_mountlist()

 	out = []
 	mounts = libgtop.glibtop_get_mountlist_l(server, byref(info), 0)
	for i in range(info.number):
 		out.append(mounts[i].mountdir)

 	glibtop_free_r(server, mounts)
 	return out

def get_fsusage(mount):
	fs_info = glibtop_fsusage()

	libgtop.glibtop_get_fsusage_l(server, byref(fs_info), mount)
	if not fs_info.blocks: return None

	if hasattr(fs_info, 'block_size'):
		block_size = fs_info.block_size
	else:
		block_size = 512

	return fs_info.bavail, fs_info.bfree - fs_info.bavail, \
		fs_info.blocks - fs_info.bfree, fs_info.blocks, \
		block_size


def get_proclist(all, uid):
	info = glibtop_proclist()
	retval = []

	if all:
		which = GLIBTOP_KERN_PROC_ALL
	else:
		which = GLIBTOP_KERN_PROC_UID
	pids = libgtop.glibtop_get_proclist_l(server, byref(info),
					c_longlong(which), c_longlong(uid))
	if pids is None:
		raise Exception('glibtop_get_proclist_l failed')
	for i in range(info.number):
		retval.append(pids[i])
	glibtop_free_r(server, pids)

	return retval

def get_proc_mem(pid):
	mem = glibtop_proc_mem()

	libgtop.glibtop_get_proc_mem_l(server, byref(mem), pid)

	if not mem.flags: raise Exception('No flags')

	rss = shared = size = 0

	if mem.flags & (1 << GLIBTOP_PROC_MEM_RSS): rss = mem.rss
	if mem.flags & (1 << GLIBTOP_PROC_MEM_SHARE): shared = mem.share
	if mem.flags & (1 << GLIBTOP_PROC_MEM_SIZE): size = mem.size

	return (rss, shared, size)

def get_proc_ppid(pid):
	uid = glibtop_proc_uid()

	libgtop.glibtop_get_proc_uid_l(server, byref(uid), pid)
	if not uid.flags: raise Exception('No flags')

	if uid.flags & (1 << GLIBTOP_PROC_UID_PPID): 
		return uid.ppid
	return 0

def get_proc_time(pid):
	info = glibtop_proc_time()

	libgtop.glibtop_get_proc_time_l(server, byref(info), pid)
	return info.utime + info.stime		# rtime not working?

def p_to_str(base, x = 0):
	# XXX: Optimise me!
	out = ""
	while base[x] != '\0':
		out = out + base[x]
		x += 1
	return out

def get_proc_state(pid):
	info = glibtop_proc_state()
	buf = glibtop_proc_args()
	
	info.state = 0

	libgtop.glibtop_get_proc_state_l(server, byref(info), pid)
	if not info.flags: raise Exception('No flags')

	if info.flags & (1 << GLIBTOP_PROC_STATE_STATE):
		if version==2:
			if info.state & GLIBTOP_PROCESS_RUNNING:
				state = 'R'
			elif info.state & GLIBTOP_PROCESS_STOPPED:
				state = 'T'
			elif info.state & GLIBTOP_PROCESS_ZOMBIE:
				state = 'Z'
			elif info.state & GLIBTOP_PROCESS_UNINTERRUPTIBLE:
				state = 'D'
			else:
				state = 'S'
		else:
			state = info.state
	else:
		state = '?'

	args = libgtop.glibtop_get_proc_args_l(server, byref(buf), pid, c_uint(0))

	if args:
		command = p_to_str(args)

	if args and len(command):
		if os.path.basename(command) == 'python':
			command = p_to_str(args, len(command) + 1)
		glibtop_free_r(server, args)
	elif info.flags & (1 << GLIBTOP_PROC_STATE_CMD):
		command = info.cmd
	else:
		command = '<unknown>'

	if info.flags & (1 << GLIBTOP_PROC_STATE_UID):
		uid = info.uid
	else:
		uid = -1

	return state, uid, command
